{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE ViewPatterns #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Engine.EngineM
( EngineM
, checkExperimentCache
, EngineEnv(..)
, EngineLog(..)
, EngineState
, mkInitialEngineState
, runEngineM
, lookupExperimentResult
, storeExperimentResult
, storeExperimentHValue
, tellLog
, nextRandomR
, profileFunction
, getHValue
) where

import Control.Category
import Control.Monad.RWS.Strict hiding (liftIO)
import Data.Int
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq, (<|))
import Debug.Trace
import Engine.EngineHValue
import Engine.ExperimentCache (ExperimentCache, Strategy)
import Engine.ExperimentResult
import Engine.Profile
import Exception
import GHC
import HscTypes
import Linker (extendLinkEnv)
import MonadUtils
import Plan
import Prelude hiding ((.), mod, id)
import qualified Data.Sequence as Seq
import qualified Engine.ExperimentCache as EC
import qualified Util.UniqMap as UniqMap
import System.Random
import UniqSupply
import Util.FCLabels hiding (get)
import Util.GHC(rawGetHValue)
import Util.UniqMap (UniqMap)
import Var

newtype EngineM a = EngineM { _runEngineM :: RWST EngineEnv EngineLog EngineState Ghc a }
  deriving (Monad, MonadReader EngineEnv, Functor) 

newtype EngineLog   = EngineLog (Seq String)
  deriving Monoid

data EngineState = EngineState { experimentCache :: !(Maybe ExperimentCache)
                               , randomGen       :: !StdGen
                               , nameHVals       :: !(UniqMap Id (First HValue))
                               }

data EngineEnv = EngineEnv { maxExpressionTime   :: !Int
                           , maxExpressionAllocs :: !Int64
                           , profile             :: !Profile
                           }

mkLabels [''EngineState, ''EngineEnv]

instance GhcMonad EngineM where
  getSession = EngineM $ lift getSession
  setSession = EngineM . lift . setSession

instance MonadUtils.MonadIO EngineM where
  liftIO = EngineM . lift . liftIO

instance ExceptionMonad EngineM where
  gcatch (EngineM c) ((_runEngineM .) -> f) = EngineM $ do
    r <- ask
    s <- get
    let f' = ($ s) . ($ r) . runRWST . f
    mapRWST (\i -> gcatch i f') c

  gblock (EngineM c) = EngineM $ mapRWST gblock c
  gunblock (EngineM c) = EngineM $ mapRWST gunblock c

instance WarnLogMonad EngineM where
  setWarnings messages = EngineM $ lift (setWarnings messages)
  getWarnings = EngineM $ lift getWarnings

instance MonadThings EngineM where
  lookupThing = EngineM . lift . lookupThing


mkInitialEngineState :: Maybe Strategy -> EngineState
mkInitialEngineState mCacheStrategy = EngineState { experimentCache = EC.newCache <$> mCacheStrategy
                                                  , randomGen = mkStdGen 0xbadbabee
                                                  , nameHVals = mempty
                                                  }

tellLog :: String -> EngineM ()
tellLog = EngineM . tell . EngineLog . Seq.singleton

lookupExperimentResult :: Experiment -> EngineM (Maybe ExperimentResult)
lookupExperimentResult !exp = profileFunction "lookupExperimentResult" $ EngineM $ 
  gets (EC.lookupExperimentResult exp <=< experimentCache)

checkExperimentCache :: Experiment -> EngineM (Maybe EngineHValue)
checkExperimentCache !exp = profileFunction "checkExperimentCache" $ EngineM $ 
  gets (EC.lookupExperimentHValue exp <=< experimentCache)

storeExperimentResult :: Experiment -> ExperimentResult -> EngineM ()
storeExperimentResult exp res = profileFunction "storeExperimentResult" $ EngineM $
  modify (mod lExperimentCache (EC.storeExperimentResult exp res <$>))

storeExperimentHValue :: Experiment -> EngineHValue -> EngineM (Maybe (Seq Experiment))
storeExperimentHValue exp hval = profileFunction "storeEngineHValue" $ do
  EngineM $ do
    !mEC <- gets (EC.lookupHValueExperiments hval <=< experimentCache)
    modify (mod lExperimentCache (EC.storeEngineHValue exp hval <$>))
    return mEC

runEngineM :: EngineState -> EngineEnv -> EngineM a -> Ghc (a, EngineState, EngineLog)
runEngineM st env code = runRWST ( _runEngineM $ (profileFunction "runEngineM" code)) env st

getHValue :: Id -> EngineM HValue
getHValue id = EngineM $ do
  First mHV <- gets (flip UniqMap.lookup id . nameHVals)
  case mHV of
    Nothing -> do
      hv <- lift $ rawGetHValue id
      modify (mod lNameHVals (\mp -> UniqMap.insert mp id (First $ Just hv)))
      return hv
    Just hv -> return hv

nextRandomR :: Int -> Int -> EngineM Int
nextRandomR lo hi = EngineM $ do
  st <- get 
  let (a, g) = randomR (lo,hi) (randomGen st)
  put $ st { randomGen = g }
  return a

profileFunction :: String -> EngineM a -> EngineM a
profileFunction name code = do
  !profile <- EngineM $ asks profile 
  recordProfile profile name code
