{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}
module Engine.Experiment
( ExperimentResult(..)
, runExperiment
, remapExperimentResult
) where

import Bottom
import Control.Applicative
import Control.Concurrent
import Control.Exception
import Control.Monad
import Control.Monad.Instances
import Control.Monad.Reader (ask)
import CoreSyn
import CoreUtils
import DataCon
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Engine.EngineHValue
import Engine.EngineM
import Engine.Eval
import Engine.ExperimentResult
import Engine.IrulanRTError
import FastString
import Format
import GHC (HValue)
import GHC.Prim
import Id
import MkCore
import MonadUtils (liftIO)
import Plan
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Traversable as T
import System.IO.Unsafe
import System.Time
import TysPrim (anyPrimTy)
import TysWiredIn (unitDataConId)
import Util.GHC

instance Format ExperimentResult where
  format (Redundant exps) = "# == " ++ format exps
  format Ok              = "."
  format (Errors exc)    = "! " ++ show exc
  format (Bottoms bid)   = "?" ++ format bid
  format (Property p)
    | p                  = "True"
    | otherwise          = "False"

runExperiment :: Experiment -> EngineM RanExperiment
runExperiment experiment = profileFunction "runExperiment" $ do
  mExistingResult <- lookupExperimentResult experiment
  
  case mExistingResult of
    Just existingResult -> do
      CachedResult existingResult <$> checkExperimentCache experiment
    Nothing -> do
      env <- ask
      val <- mkExperimentValue experiment

      resultVar <- liftIO $ newEmptyMVar
      evalThreadId <- liftIO $ forkIO (evalWithSeq val >>= ((putMVar resultVar) $!))

      evalResult <- liftIO $ watchMemoryAndTime env resultVar

      liftIO (killThread evalThreadId)
      liftIO getAllocations >>= \allocs -> tellLog ("Allocations: " ++ show allocs)

      let simpleCase result hv = storeExperimentResult experiment result >> return (result, hv)

      (!ranExp, !mEHVal) <- case evalResult of
        EvalException (fromException -> Just (Bottom i)) -> simpleCase (Bottoms i) Nothing
        EvalException e  -> simpleCase (Errors e) Nothing
        EvalOk val -> do
          if experimentIsProperty experiment
            then simpleCase (Property (unsafeCoerce# (toHValue val))) (Just val)
            else do
              if experimentHasBottom experiment
                then simpleCase Ok (Just val)
                else (maybe (Ok, Just val)) ((,Just val) . Redundant) <$> storeExperimentHValue experiment val
      return $! NewResult ranExp mEHVal

watchMemoryAndTime :: EngineEnv -> MVar EvalResult -> IO EvalResult
watchMemoryAndTime env er = do
  startTime <- getClockTime
  let endTime = noTimeDiff { tdPicosec = 1000000 * (fromIntegral (maxExpressionTime env)) } `addToClockTime` startTime 
  baseAllocs <- getAllocations
  let endAllocs = baseAllocs + maxExpressionAllocs env

  let loop = do
              newAllocs <- getAllocations
              newTime   <- getClockTime
              mVal      <- tryTakeMVar er

              let go
                    | Just res  <- mVal
                        = return res
                    | newAllocs > endAllocs 
                        = return (EvalException $! SomeException $! ErrorCall "<allocated too much memory!>")
                    | newTime > endTime
                        = return (EvalException $! SomeException $! ErrorCall "<ran out of time!>")
                    | otherwise = yield >> loop
              go
    
  loop

getOrMkExperimentValue :: Experiment -> EngineM CompiledExpression
getOrMkExperimentValue exp
  | not (experimentHasBottom exp) = profileFunction "getOrMkExperimentValue" $ (checkExperimentCache exp >>= maybe (mkExperimentValue exp)
                                                                                                                   (return . liftM2 CE toHValue toType))
  | otherwise = profileFunction "getOrMkExperimentValue" $ mkExperimentValue exp

mkExperimentValue :: Experiment -> EngineM CompiledExpression
mkExperimentValue !experiment = profileFunction "mkExperimentValue" $ do
  rootHVal <- mkExperimentRootValue
  childHVals <- catMaybes . F.toList <$> T.mapM mkArgumentValue (experimentArgs experiment)
  return $! CE (applyHValues rootHVal childHVals) (experimentType experiment)
  where
    mkExperimentRootValue :: EngineM HValue
    mkExperimentRootValue
      | ConstantRoot c <- experimentRoot experiment = return $! mkConstantValue c
      | IdRoot x <- experimentRoot experiment = getHValue x
      | DestructRoot { destructCaseOf 
                     , destructConstructor 
                     , destructVarIndex 
                     , destructResultType 
                     } <- experimentRoot experiment = do
        
      (CE hv _) <- getOrMkExperimentValue destructCaseOf
      return $ unsafeCoerce# (unsafePerformIO $ getConstructorArgument hv destructVarIndex)

    mkArgumentValue :: ExperimentArg -> EngineM (Maybe HValue)
    mkArgumentValue (ArgExperiment exp) = (\(CE hv _) -> Just hv) <$> getOrMkExperimentValue exp
    mkArgumentValue (ArgBottom _ bid)   = return $ Just (unsafeCoerce# (throw (Bottom bid)))
    mkArgumentValue (ArgType ty)        = return $ Nothing
    mkArgumentValue ArgUnconstrained    = return $ Just (unsafeCoerce# ())
    
    applyHValues :: HValue -> [HValue] -> HValue
    applyHValues v [] = v
    applyHValues v (x:xs) = applyHValues ((unsafeCoerce# v) x) xs

    mkConstantValue :: Constant -> HValue
    mkConstantValue (IntConstant c) = unsafeCoerce# c
    mkConstantValue (CharConstant c) = unsafeCoerce# c
    mkConstantValue (FloatConstant c) = unsafeCoerce# c
    mkConstantValue (DoubleConstant c) = unsafeCoerce# c
    mkConstantValue (IntegerConstant c _) = unsafeCoerce# c
