{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE BangPatterns #-}
module Engine.PlanStrategyTreePlan
( DepthLimitedRandomFork(..)
, DFSStrategy(..)
, DFSEvenTime(..)
, RestartingRandom(..)
) where

import Control.Applicative
import Control.Monad (guard)
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Data.Sequence ((><), Seq)
import Data.Time.Clock
import Engine.EngineHValue
import Engine.EngineM
import Engine.Experiment
import Engine.ExperimentResult
import Engine.RuntimePlan
import Format
import MonadUtils
import Plan
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Util.BiMap as BiMap
import qualified Util.UniqMap as UniqMap
import Util
import Util.BiMap (BiMap)
import Util.UniqMap (UniqMap)

data DFSStrategy = DFSS { dfsGetTreePlan    :: TreePlan
                        , dfsDepthRemaining :: !Depth
                        , dfsExpDepthLimit  :: !(Maybe Int)
                        , dfsExpSizeLimit   :: !(Maybe Int)
                        }

instance PlanStrategy DFSStrategy where
  select options = return 0

  update d@DFSS { dfsGetTreePlan, dfsDepthRemaining , dfsExpDepthLimit , dfsExpSizeLimit }
    | dfsDepthRemaining <= 0 
      = return (Seq.empty, Nothing)
    | TPStep { tpExperiment } <- dfsGetTreePlan
    , Just expDepthLimit <- dfsExpDepthLimit
    , experimentMaxDepth tpExperiment > expDepthLimit 
      = return (Seq.empty, Nothing)
    | TPStep { tpExperiment } <- dfsGetTreePlan
    , Just expSizeLimit <- dfsExpSizeLimit
    , experimentSize tpExperiment > expSizeLimit 
      = return (Seq.empty, Nothing)
    | TPStep { tpExperiment } <- dfsGetTreePlan
      = updateFn dfsGetTreePlan tpExperiment mkNext
    | TPSplit { tpChildren } <- dfsGetTreePlan 
      = return (mkNext <$> tpChildren, Nothing)
    where
      mkNext kid = d { dfsGetTreePlan = kid 
                     , dfsDepthRemaining = dfsDepthRemaining - 1}

newtype DepthLimitedRandomFork = DLRF { dlrfDFSS :: DFSStrategy }

instance PlanStrategy DepthLimitedRandomFork where
  select options = nextRandomR 0 (Seq.length options - 1)

  update d@DLRF { dlrfDFSS } = do
    (newChildren, mResult) <- update dlrfDFSS
    let next = DLRF <$> newChildren
    return (next, mResult)

data RestartingRandom = RR { rrGetRoot    :: !(() -> TreePlan)
                           , rrDepthLimit :: !Depth
                           , rrDFSPart    :: !DFSStrategy
                           }

instance PlanStrategy RestartingRandom where
  select options = return 0

  update r@RR { rrGetRoot, rrDepthLimit, rrDFSPart } = do
    (newChildren, mResult) <- update rrDFSPart

    next <- if Seq.null newChildren 
              then return $ Seq.singleton $ RR { rrGetRoot = rrGetRoot
                                               , rrDFSPart = rrDFSPart { dfsDepthRemaining = rrDepthLimit 
                                                                       , dfsGetTreePlan    = rrGetRoot () 
                                                                       }
                                               , rrDepthLimit = rrDepthLimit
                                               }
              else do choice <- nextRandomR 0 (Seq.length newChildren - 1)
                      return $ Seq.singleton $ r { rrDFSPart = Seq.index newChildren choice }
    return (next , mResult)



data DFSEvenTime = DFSET { dfsetFinishTime :: UTCTime
                         , dfsetDFSPart    :: DFSStrategy
                         }

instance PlanStrategy DFSEvenTime where
  select options = return 0

  update d@DFSET { dfsetFinishTime , dfsetDFSPart } = do
    
    timeNow <- liftIO $ getCurrentTime

    if timeNow > dfsetFinishTime 
      then return (Seq.empty, Nothing)
      else do
        (newChildren, mResult) <- update dfsetDFSPart
        let timeIncr = (dfsetFinishTime `diffUTCTime` timeNow) / (fromIntegral (Seq.length newChildren))
        let incrF = (timeIncr `addUTCTime`)
        let kids = Seq.zipWith DFSET (Seq.fromList . take (Seq.length newChildren) . iterate incrF $ (incrF timeNow)) newChildren
        return (kids, mResult)
        


updateFn :: TreePlan -> Experiment -> (TreePlan -> a) -> EngineM (Seq.Seq a , Maybe (Experiment, RanExperiment))
updateFn currentPlan !experiment mkChild = profileFunction "updateFn" $ do
  --() <- mTrace $ "running: " ++ format experiment
  ranExperiment@(getExperimentResult -> result) <- (runExperiment experiment)

  let okLogic = profileFunction "updateFn:okLogic" $ do
                  let okPs = tpOnOk currentPlan 
                  casePs <- case experimentRTValue ranExperiment of
                                 Nothing -> return $ mempty
                                 Just hv -> do
                                   let mp = tpOnDataCon currentPlan
                                   mv <- if (UniqMap.null mp) 
                                           then return Nothing
                                           else getConstructor hv
                                   return $! maybe mempty (UniqMap.lookup mp) mv
                  return $ (okPs `mappend` casePs)
  let bottomLogic bid = profileFunction "updateFn:bottomLogic" $
                          return $! fromMaybe (error "PlanStrategy updateFn onBottom") . Map.lookup bid . tpOnBottom $ currentPlan
                      
  !next <- case result of
            Ok          -> okLogic
            Bottoms bid -> bottomLogic bid 
            Property _  -> okLogic
            _           -> return $ mempty
    
  return $ (Seq.singleton $ mkChild next, Just (experiment, ranExperiment))
