{-# LANGUAGE TupleSections #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE BangPatterns #-}
module IdDB.ConstructorGraph
( empty
, addRoot
, destructPlan
, step
, accepting
, ConstructorGraph(..)
, DestructPlan
) where

import Control.Arrow
import Control.Monad.RWS.Strict
import DataCon
import Data.Functor
import Data.Graph.Inductive (Node)
import Data.Graph.Inductive.PatriciaTree (Gr)
import Data.IntMap (IntMap)
import Data.IntSet (IntSet)
import Data.List
import Data.Maybe
import Format
import IdDB.FixedTypeMap (FixedTypeMap)
import IdDB.TypeMap (TypeMapInt)
import Prelude hiding (mod)
import qualified Data.Graph.Inductive as Gr
import qualified Data.IntMap as IntMap
import qualified Data.IntSet as IntSet
import qualified IdDB.FixedTypeMap as FTMap
import qualified IdDB.TypeMap as Map
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import TyCon
import Type
import Outputable (showPpr)
import IdDB.TypeMapDebug
import Util
import Util.FCLabels
import Util.GHC
import Util.UniqMap (UniqMap)
import Util.UniqSet (UniqSet)

data ConstructorGraph x
  = CG { knownDataCons  :: !(UniqMap TyCon (UniqSet DataCon))
       , depthLimit     :: !Integer
       , theGraph       :: !(Gr (Type, Integer) [(DataCon, Int)])
       , typeNodes      :: !TypeMapInt
       , fixedTypeNodes :: !(FixedTypeMap Node)
       , rootNodes      :: !(IntMap [x]) 
       , nextNode       :: !Node
       }

data DestructPlan 
  = DestructPlan
  { accepting :: !Bool
  , step      :: !(Maybe (UniqMap DataCon (IntMap DestructPlan)))
  }

newtype BuildM x a = BuildM { runBuildM :: RWS BuildEnv () (BuildState x) a }
  deriving Monad

data BuildEnv
  = BuildEnv 
  { beDepthRemaining :: !Integer
  , beKnownDataCons  :: !(UniqMap TyCon (UniqSet DataCon))
  }

data BuildState x
  = BuildState
  { bsTheGraph       :: !(Gr (Type, Integer) [(DataCon, Int)])
  , bsTypeNodes      :: !TypeMapInt
  , bsFixedTypeNodes :: !(FixedTypeMap Node)
  , bsNextNode       :: !Node
  }

mkLabels [''BuildEnv, ''BuildState]

empty :: UniqMap TyCon (UniqSet DataCon) -> Integer -> ConstructorGraph a
empty knownDataCons depthLimit 
  = CG { knownDataCons  = knownDataCons
       , depthLimit     = depthLimit
       , theGraph       = Gr.empty
       , typeNodes      = Map.empty
       , fixedTypeNodes = FTMap.empty
       , rootNodes      = IntMap.empty
       , nextNode       = 0
       }

addRoot :: ConstructorGraph x -> (x, Type) -> ConstructorGraph x
addRoot cg@CG { knownDataCons, depthLimit, theGraph
              , typeNodes, fixedTypeNodes, rootNodes, nextNode 
              } 
        (root, rootsType) 
  = cg { theGraph       = bsTheGraph endState
       , typeNodes      = bsTypeNodes endState
       , fixedTypeNodes = bsFixedTypeNodes endState
       , rootNodes      = IntMap.insertWith (++) rootNode [root] $ rootNodes
       , nextNode       = bsNextNode endState
       }
  where
    Just rootNode = FTMap.lookup rootsType' (bsFixedTypeNodes endState)

    rootsType' = last . listFunTys . dropForAlls $ rootsType

    (_, endState, _) = run initialEnv initialState . runBuildM $ buildChain rootsType'
      where
        initialEnv = BuildEnv { beDepthRemaining = depthLimit
                              , beKnownDataCons  = knownDataCons
                              }

        initialState = BuildState { bsTheGraph       = theGraph
                                  , bsTypeNodes      = typeNodes
                                  , bsFixedTypeNodes = fixedTypeNodes
                                  , bsNextNode       = nextNode
                                  }
    run r s code = runRWS code r s
    
buildChain :: Type -> BuildM x ()
buildChain ty = do
  (tyNodeId,_) <- getUpdateOrAddNode ty
  nexts <- nextTypes ty
  forM_ nexts $ \(ty', dcI) -> do
    (ty'NodeId, isNewer) <- getUpdateOrAddNode ty'
    addEdgeIfNotPresent ty'NodeId tyNodeId dcI
    when isNewer (increaseDepth $ buildChain ty')

getUpdateOrAddNode :: Type -> BuildM x (Node, Bool)
getUpdateOrAddNode ty = BuildM $ do
  found <- gets (FTMap.lookup ty . bsFixedTypeNodes)

  clim <- asks beDepthRemaining

  case found of
    Nothing       -> do
      tyNodeId <- gets bsNextNode
      modify $ mod lBsNextNode succ
             . mod lBsTheGraph (Gr.insNode (tyNodeId, (ty, clim)))
             . mod lBsTypeNodes (\tm -> Map.addIntProvider tm ty tyNodeId)
             . mod lBsFixedTypeNodes (\tm -> FTMap.insert tm ty tyNodeId)
      return (tyNodeId, True)

    Just tyNodeId -> do
      gr <- gets bsTheGraph
      let (_, lim) = fromJust $ Gr.lab gr tyNodeId

      let isNewer = clim > lim
      when isNewer $ modify (mod lBsTheGraph (Gr.insNode (tyNodeId, (ty, clim))))
        
      return (tyNodeId, isNewer)

nextTypes :: Type -> BuildM x [(Type, (DataCon, Int))]
nextTypes ty = BuildM (asks $ nextTypes' . beKnownDataCons)
  where
    nextTypes' :: UniqMap TyCon (UniqSet DataCon) -> [(Type, (DataCon, Int))]
    nextTypes' knownDataCons 
      | Just (leadTyCon, appTys) <- splitTyConApp_maybe ty = do
          dc <- UniqSet.toList $ UniqMap.lookup knownDataCons leadTyCon
          (index, argTy) <- [0..] `zip` (dataConInstArgTys dc appTys)
          return (argTy, (dc, index))
      | otherwise = []
    
increaseDepth :: BuildM x () -> BuildM x ()
increaseDepth (BuildM code) = BuildM $ do
  dr <- asks beDepthRemaining
  if dr > 1 
    then (local (mod lBeDepthRemaining pred) code)
    else return ()

addEdgeIfNotPresent :: Node -> Node -> (DataCon, Int) -> BuildM x ()
addEdgeIfNotPresent fromNode toNode value = BuildM $ do
  gr <- gets bsTheGraph
  case lookup toNode . Gr.lsuc gr $ fromNode of
    Nothing -> modify (mod lBsTheGraph (Gr.insEdge (fromNode, toNode, [value])))
    Just v  -> modify (mod lBsTheGraph (Gr.insEdge (fromNode, toNode, (value : (filter (/= value) v)))
                                       . Gr.delEdge (fromNode, toNode)))

destructPlan :: Type -> ConstructorGraph x -> [((x, TvSubst), DestructPlan)]
destructPlan ty cg@CG { theGraph, typeNodes , rootNodes }   = do
  endNode <- UniqSet.toList . mconcat . map snd . Map.lookupProviders ty $ typeNodes

  let reachableNodes = Gr.reachable endNode theGraph
  let nodesToRemove = Gr.nodes theGraph \\ reachableNodes

  let thisGraph = Gr.grev . Gr.delNodes nodesToRemove $ theGraph

  (rootNode, xs) <- mapMaybe (\rn -> (rn,) <$> IntMap.lookup rn rootNodes) reachableNodes
  let hasChildren = not . null $ Gr.lsuc thisGraph rootNode
  guard (rootNode /= endNode || hasChildren)

  x <- xs

  let Just (endType, _) = Gr.lab thisGraph endNode
  let Just subst        = unifyTypes ty endType
  return ((x, subst), mkDestructPlan False thisGraph rootNode endNode)
  where
    mkDestructPlan :: Bool -> Gr (Type, Integer) [(DataCon, Int)] -> Node -> Node -> DestructPlan
    mkDestructPlan canAccept thisGraph currentNode endNode 
      = DestructPlan { accepting = canAccept && currentNode == endNode
                     , step = theStep
                     }
      where
        next = Gr.lsuc thisGraph currentNode

        theStep
          | null next = Nothing
          | otherwise = Just $ mconcat . map mkMapping $ next

        mkMapping :: (Node, [(DataCon, Int)]) -> UniqMap DataCon (IntMap DestructPlan)
        mkMapping (nextNode, possibles) = mconcat . map (uncurry UniqMap.singleton . (second $ flip IntMap.singleton nextDP)) $ possibles
          where
            nextDP = mkDestructPlan True thisGraph nextNode endNode
