{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE UnicodeSyntax #-}
module Vis.GL() where

import Bottom
import DataCon
import Control.Applicative
import Data.List
import Data.Maybe
import Engine
import Format
import GraphRewriting.GL.Render
import GraphRewriting.Layout.PortSpec
import Outputable
import Plan
import Prelude hiding (exp)
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import qualified Graphics.UI.GLUT as GL
import qualified Util.BiMap as BiMap
import Vis.Graph

drawPorts ∷ PortSpec n ⇒ n → IO ()
drawPorts = mapM_ drawPort . relPortPos

circle r1 r2 step = mapM_ vertex2 vs where
	is = take (truncate step + 1) [0, i' .. ]
	i' = 2 * pi / step
	vs = [ (r1 * cos i, r2 * sin i) | i <- is ]

drawPort pos = GL.preservingMatrix $ do
	GL.translate $ vector pos
	GL.renderPrimitive GL.Polygon (circle 0.15 0.15 10)

drawNode label = do
	GL.renderPrimitive GL.LineLoop (circle 1 1 20)
	drawString label

drawString label = GL.preservingMatrix $ do
	GL.translate $ vector2 (-0.3,-0.3)
	GL.scale 0.0065 0.0065 (0 ∷ GL.GLdouble)
	GL.renderString GL.MonoRoman label

instance PortSpec PlanNode where
  portSpec node = let sd = sameDir in case node of
    Root {} -> [sd s]
    RanExperiment {} -> [sd n, sd s]
    Split { children } -> (sd n) : [ sd ((sw `mult` (i/l)) + (se `mult` (1 - (i/l)))) |  i <- [0 .. l - 1]] where l = genericLength children
    PlanToDo {} -> [sd n]
    Nil {} -> [sd n]
    where -- points of compass
            n = Vector2 0 1
            w = Vector2 (-1) 0
            e = Vector2 1 0
            s = Vector2 0 (-1)
            sw = Vector2 (-1) (-1)
            se = Vector2 1 (-1)
            nw = Vector2 (-1) 1
            ne = Vector2 1 1
            Vector2 x y `mult` z = Vector2 (x * z) (y * z)

instance Render PlanNode where 
  render node = drawPorts node >> case node of
    Root {}                       -> return ()
    RanExperiment { exp, result } -> drawNode (simpleOutput exp result)
    Split {}                      -> drawNode "<>"
    PlanToDo { plan }             -> case plan of TPStep { tpExperiment } -> drawNode (simpleOutputExp tpExperiment)
                                                  TPSplit { tpChildren } -> drawNode ("[ " ++ show (Seq.length tpChildren) ++ "]")
    Nil {}                        -> return ()
    

simpleOutputExp :: Experiment -> String
simpleOutputExp exp' = simpleOutput' exp 
  where
    (exp,_) = canonicaliseExperiment exp'

simpleOutput :: Experiment -> ExperimentResult -> String
simpleOutput exp' res' = simpleOutput' exp ++ arrow ++ simpleOutputResult res
  where
    arrow = " ==> "

    (exp, bimap) = canonicaliseExperiment exp'
    res = remapExperimentResult (BiMap.forwardMap bimap) res'

simpleOutput' exp 
  = unwords $ [simpleOutputRoot (not . Seq.null $ experimentArgs exp) (experimentRoot exp)] ++ (catMaybes . F.toList $ simpleOutputArg <$> experimentArgs exp) 

simpleOutputArg (ArgExperiment e)
  | DestructRoot {} <- experimentRoot e = Just $ "(" ++ simpleOutput' e ++ ")"
  | Seq.null (experimentArgs e)         = Just $ simpleOutput' e
  | otherwise                           = Just $ "(" ++ simpleOutput' e ++ ")"
simpleOutputArg (ArgBottom _ b)  = Just $ simpleBottom b
simpleOutputArg (ArgType _)      = Nothing
simpleOutputArg ArgUnconstrained = Just $ "*"

simpleOutputRoot _ (ConstantRoot x) = format x
simpleOutputRoot _ (IdRoot x) = showSDocUnqual . ppr $ x
simpleOutputRoot b (DestructRoot { destructCaseOf   
                               , destructConstructor
                               , destructVarIndex }) 
  | b = "(" ++ mainBit ++ ")"
  | otherwise = mainBit
  where
    mainBit = "case " ++ (simpleOutput' destructCaseOf) ++ " of " ++ dcBit ++ " -> x"
    dcBit
      | dataConIsInfix destructConstructor = head args ++ " " ++ dcString ++ " " ++ unwords (tail args)
      | otherwise = dcString ++ " " ++ unwords args
      where
        dcString = showSDocUnqual . ppr $ destructConstructor
        args = [ if y == destructVarIndex then "x" else "_" | y <- [0 ..(length $ dataConOrigArgTys destructConstructor) - 1]]

simpleBottom b
  | b == initialBottomId         = "?"
  | otherwise                    = "?" ++ format b

simpleOutputResult (Bottoms b) = simpleBottom b
simpleOutputResult r           = format r
