{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE ViewPatterns #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module TestRunner(run) where

import Control.Applicative hiding ((<|>))
import Control.Monad
import Control.Monad.State.Strict
import Data.List
import Data.Map (Map)
import Data.Maybe
import Data.Word
import Format
import qualified Data.Map as Map
import System.Exit
import System.FilePath
import System.IO
import System.Posix.Files
import System.Process
import Test

data TestResult = TROk
                | TRErr { trStdOut :: FileResult 
                        , trStdErr :: FileResult
                        }
  deriving Show

data FileResult = Ok
                | Diff String
                | NoExpectedFile String
  deriving Show

run :: FilePath -> [Test] -> IO ()
run = runTests

formatResult :: Test -> TestResult -> String
formatResult _ TROk = ""
formatResult t TRErr { trStdOut , trStdErr } = modu t ++ ":" ++ "\n"
                                            ++ formatFileResult "stdout" trStdOut
                                            ++ formatFileResult "stderr" trStdErr
  where
    formatFileResult :: String -> FileResult -> String
    formatFileResult _ Ok = ""
    formatFileResult name (Diff x)           = "  " ++ redWrap (name ++ ":\n") ++ x
    formatFileResult name (NoExpectedFile s) = "  " ++ redWrap (name ++ " no expected output, got:\n") ++ s
    
redWrap x = "\ESC[31m" ++ x ++ "\ESC[0m"
greenWrap x = "\ESC[32m" ++ x ++ "\ESC[0m"

runTests :: FilePath -> [Test] -> IO ()
runTests testDir tests = do
  summary <- forM tests $ \test -> do
    res <- runTest testDir test
    let resStr = formatResult test res
    unless (null resStr) (putStrLn resStr)
    case res of
      TROk -> return True
      _    -> return False
  
  let totalTests = length summary
  let passes     = length . filter id $ summary
  let fails      = totalTests - passes

  let wrap
        | fails == 0 = greenWrap
        | otherwise = redWrap
  putStrLn $ wrap (unwords ["Total:", show totalTests, "Passes:", show passes, "Fails:", show fails ])


runTest :: FilePath -> Test -> IO TestResult
runTest srcDir test = do
  let process = proc "./irulan" (  preOptions test 
                                ++ [cmd test]
                                ++ postOptions test
                                ++ ["--ghc-options=-i" ++ srcDir]
                                ++ [modu test]
                                )
 
  let actOut = srcDir </> (modu test) <.> "out" <.> "actual"
  let actErr = srcDir </> (modu test) <.> "err" <.> "actual"

  outHandle <- openFile actOut WriteMode 
  hSetBuffering outHandle NoBuffering
  errHandle <- openFile actErr WriteMode 
  hSetBuffering errHandle NoBuffering

  (_, _, _, phdl) <- createProcess process { std_out = UseHandle outHandle
                                           , std_err = UseHandle errHandle
                                           }
  waitForProcess phdl

  hClose outHandle
  hClose errHandle

  let expOut = srcDir </> (modu test) <.> "out"
  let expErr = srcDir </> (modu test) <.> "err"

  outRes <- checkTwoFiles expOut actOut
  errRes <- checkTwoFiles expErr actErr

  return $ 
    case (outRes, errRes) of
      (Ok, Ok) -> TROk
      _        -> TRErr { trStdOut = outRes
                        , trStdErr = errRes
                        }


checkTwoFiles :: FilePath -> FilePath -> IO FileResult
checkTwoFiles expected actual = do
  eExists <- fileExist expected 

  if eExists
    then do
      (exitCode, stdout, stderr) <- readProcessWithExitCode "diff" [expected, actual] []
      case exitCode of
        ExitSuccess   -> return Ok
        ExitFailure 1 -> return $ Diff stdout
        ExitFailure n -> error ("Diff didn't return 0 or 1!: " ++ show stderr)
    else do
      actualContents <- readFile actual
      if null actualContents
        then return Ok
        else return $ NoExpectedFile actualContents
