{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Plan.TestLib where

import Control.Arrow
import Control.Monad
import Control.Monad.Error hiding (liftIO)
import Control.Monad.State hiding (liftIO)
import Convert
import Data.Either
import Data.Graph.Inductive as G
import Data.List
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Data.Ord
import Data.Sequence (Seq)
import Data.Set (Set)
import Digraph
import FastString
import Format
import GHC hiding (Failed, Succeeded)
import GHC.Paths
import HscTypes
import Id
import IdDB
import Language.Haskell.TH (TypeQ)
import MonadUtils
import Outputable hiding (Depth)
import Plan
import Plan.Debug
import Plan.Types hiding (Name)
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Data.Set as Set
import qualified Data.Traversable as T
import qualified Language.Haskell.TH as TH
import qualified Plan.Types
import qualified Test.HUnit as HUnit
import System.FilePath
import TcSplice
import Test.HUnit ((~:), (@?))
import Type
import TysWiredIn (unitTy)
import Unique
import Util
import Util.GHC
import Util.TH
import VarEnv


infixr 9 |>

(|>) :: TypeQ -> TypeQ -> TypeQ
x |> y = liftM2 (\x y -> TH.AppT (TH.AppT TH.ArrowT x) y) x y

(-->) = (,)

data Test 
  = Test
  { modu                 :: String
  , rootFunction         :: TH.Name
  , expectedIdsForTypes  :: [(TypeQ, [(TH.Name, Maybe TypeQ)])]
                                       -- The Maybe TypeQ is if a substitution is expected,
                                       -- and what the id's type should be when it is applied to the type.
  , outputDot            :: Bool
  , idSourceToUse        :: IdSource
  , treePlanMaxDepth     :: Int
  , treePlanCaseArgument :: Bool
  , treePlanCaseDestruct :: Bool
  }
  | NoTestYet

test = Test { modu                 = error "set modu!"
            , rootFunction         = error "set root function"
            , expectedIdsForTypes  = error "set expected ids for types"
            , outputDot            = False
            , idSourceToUse        = TestModule
            , treePlanMaxDepth     = 5
            , treePlanCaseArgument = False
            , treePlanCaseDestruct = False
            }

noTestYet = NoTestYet

isATest NoTestYet = False
isATest Test {}   = True


mkPlanTest :: Test -> HUnit.Test
mkPlanTest test = (modu test) ~: runGhc (Just libdir) $ do
  dflags <- getSessionDynFlags 
  setSessionDynFlags dflags
  defaultCleanupHandler dflags $ do
    guessTarget ("Plan." ++ (modu test)) Nothing >>= \target -> setTargets [target]
    result <- load LoadAllTargets

    liftIO $ (succeeded result) @? "Couldn't load module: " ++ show (modu test)

    mg <- depanal [] False
    let mods = flattenSCCs (topSortModuleGraph False mg Nothing)
    forM_ mods $ \modsum -> do
      printWarnings

      modInfo <- fromMaybe (error $ "Couldn't get module info: " ++ showPpr modsum) <$>
                           (getModuleInfo (ms_mod modsum))
      modIface <- getModIface modsum

      let (modName :: String) = moduleNameString . ms_mod_name $ modsum

      when (modName == ("Plan." ++ modu test)) $ do
        targetId <- idFromTHName (rootFunction test)
        initDB <- mkIdDBForIds [targetId] (idSourceToUse test) [] mempty

        ghcEIFTs <- ghcizeExpectedIdsForTypes (expectedIdsForTypes test)
        
        liftIO $ checkExpectedIdsForTypes ghcEIFTs initDB

        let treePlan = generateTreePlan initDB (treePlanCaseArgument test) (treePlanCaseDestruct test)
        liftIO $ checkTreePlan (modu test) (treePlanMaxDepth test) treePlan


ghcizeExpectedIdsForTypes :: [(TypeQ, [(TH.Name, Maybe TypeQ)])] -> Ghc [(Type, [(Id, Maybe Type)])]
ghcizeExpectedIdsForTypes = mapM (\(tq, namSubs) -> liftM2 (,) (typeFromTHTypeQ tq)
                                                               (mapM (uncurry ghcizeNamSubs) namSubs))
  where
    ghcizeNamSubs :: TH.Name -> Maybe TypeQ -> Ghc (Id, Maybe Type)
    ghcizeNamSubs nam mTypeQs = liftM2 (,) (idFromTHName nam) mTypes
      where
        mTypes :: Ghc (Maybe Type)
        mTypes = T.sequence (typeFromTHTypeQ <$> mTypeQs)

checkTreePlan :: String -> Int -> TreePlan -> IO ()
checkTreePlan moduName maxDepth actualPlan = do
  expectedDebugPlan <- readFile ("Plan" </> moduName <.> "treeplan")
  let actualDebugPlan = dotifyTreePlan maxDepth actualPlan 

  unless (expectedDebugPlan == actualDebugPlan) $ do
    writeFile ("Plan" </> moduName <.> "treeplan" <.> "debug" <.> "out") actualDebugPlan

  expectedDebugPlan == actualDebugPlan @? "Expected Tree Plan /= Actual Tree Plan"
  
  
checkExpectedIdsForTypes :: [(Type, [(Id, Maybe Type)])] -> IdDB -> IO ()
checkExpectedIdsForTypes expecteds db = forM_ expecteds (uncurry checkExpectedIdsForType)
  where
    checkExpectedIdsForType :: Type -> [(Id, Maybe Type)] -> IO ()
    checkExpectedIdsForType ty (sortBy (comparing fst)  -> expIdSubs) = do 
      let expIds = map fst expIdSubs
      let actIdSubs = sortBy (comparing fst3) $ (idsForType emptyInScopeSet (dropForAlls ty)) db 
      let actIds = map fst3 actIdSubs

      expIds == actIds @? "For type: " ++ show ty ++ " expected ids: " ++ show expIds ++ ", got: " ++ show actIds

      zipWithM_ checkOneExpectedId expIdSubs actIdSubs

    checkOneExpectedId :: (Id, Maybe Type) -> (Id, TvSubst, InScopeSet) -> IO ()
    checkOneExpectedId (expId, _) (actId, _, _)
      | expId /= actId = False @? "Expected id /= actual id when they really should!"
    checkOneExpectedId (_, Nothing) (_, subst, _) 
      = isEmptyTvSubst subst @? "Didn't expect a substitution, but got one!"
    checkOneExpectedId (expId, Just expType) (actId, aSubs, _) 
      = actType == substExpTy @?  "Expected type /= actual type post subst. Expected: " ++ 
                               show (dropForAlls expType) ++ ", actual: " ++ show actType ++ 
                               ", raw: " ++ show rawActType
      where
        (_, rawActType) = splitForAllTys (idType actId)
        actType = substTy aSubs rawActType

        substExpTy = (substTy aSubs $ dropForAlls expType) 
    checkOneExpectedId (expId, expType) (actId, actSubs, _) 
      = False @? "checkOneExpectedId failed for " ++ show expId ++ "{" ++ 
                 show expType ++ "}, got " ++ show actId



mkNoSubstEIFT :: [(TypeQ, [TH.Name])] -> [(TypeQ, [(TH.Name, Maybe TypeQ)])]
mkNoSubstEIFT = map (second (map (\n -> (n, Nothing))))
